# -*- coding: UTF-8 -*-
"""
  PyNOT - longslit reduction

  The module handles longslit reduction for the ALFOSC instrument
  at the Nordic Optical Telescope (NOT).
  Written by Jens-Kristian Krogager, Mar 2017

  depends on matplotlib, numpy, scipy, pyfits, argparse, astroscrappy

  .. functions:
   - combine_bias_frames
   - combine_flat_frames
   - normalize_spectral_flat
   - fit_background
   - science_reduction_2d
"""

# -- Notes for trace identification:
# Use scipy.signal.find_peaks to identify number of objects in slit:
# Construct SPSF and determine the background noise: sigma.
#
#  peaks = scipy.signal.find_peaks(SPSF, height=3*sigma, width=3.)
#
#  numbers like this seem to work fairly well, the number of sigmas should be a parameter
#  this required an object to be more than 3 pixels wide and more than 3 sigma above background
# The result is:
#  loc = peaks[0]  # and array of pixel locations of the center of each peak
#  height = peaks[1]['peak_heights']   # an array of peak height
# Then the objects will be sorted according to their 'height' i.e., their flux:
#  sorted_objects = sorted(list(zip(loc, height)), key=lambda x: x[1], reverse=True)
#  trace_centers = np.array(sorted_objects)[:, 0]   # sorted array of central pixels


__author__ = 'Jens-Kristian Krogager'
__email__ = "krogager@iap.fr"
__credits__ = ["Jens-Kristian Krogager"]

import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from scipy.ndimage import median_filter
from scipy import signal
from numpy.polynomial import Chebyshev
import os
import warnings

from astroscrappy import detect_cosmics

from pynot.alfosc import create_pixel_array, get_header
from pynot.functions import mad, get_version_number


__version__ = get_version_number()


def trim_overscan(img, hdr, overscan=50, mode='spec'):
    """Trim the overscan regions on either side in X and on top in Y"""
    # Trim overscan
    if mode == 'spec':
        X = create_pixel_array(hdr, 1)
        Y = create_pixel_array(hdr, 2)
    else:
        X = np.arange(img.shape[1]) + 1
        Y = np.arange(img.shape[0]) + 1
    img_region_x = (X >= overscan) & (X <= 2148-overscan)
    xlimits = img_region_x.nonzero()[0]
    img_region_y = (Y <= 2102-overscan)
    ylimits = img_region_y.nonzero()[0]
    x1 = min(xlimits)
    x2 = max(xlimits)+1
    y1 = min(ylimits)
    y2 = max(ylimits)+1
    img_trim = img[y1:y2, x1:x2]
    if mode == 'spec':
        hdr['CRVAL1'] += x1
        hdr['CRVAL2'] += y1

    else:
        hdr['CRPIX1'] -= x1
        hdr['CRPIX2'] -= y1
    hdr['NAXIS1'] = img_trim.shape[1]
    hdr['NAXIS2'] = img_trim.shape[0]
    return img_trim, hdr


def detect_filter_edge(fname, overscan=50):
    """Automatically detect edges in the normalized flat field"""
    # Get median profile along slit:
    img = fits.getdata(fname)
    hdr = get_header(fname)
    if 'OVERSCAN' in hdr:
        overscan = 0

    # Using the normalized flat field, the values are between 0 and 1.
    # Convert the image to a binary mask image:
    img[img > 0.5] = 1.
    img[img < 0.5] = -1.

    fx = np.median(img, 0) > 0.5
    fy = np.median(img, 1) > 0.5

    # Detect initial edges:
    x1 = np.min(fx.nonzero()[0])
    x2 = np.max(fx.nonzero()[0])

    y1 = np.min(fy.nonzero()[0])
    y2 = np.max(fy.nonzero()[0])

    # If the edge is curved, decrease the trim edges
    # until they are fully inside the image region
    lower = img[y1, x1]
    upper = img[y2, x2]
    while lower < 0 and upper < 0:
        x1 += 1
        y1 += 1
        x2 -= 1
        y2 -= 1
        lower = img[y1, x1]
        upper = img[y2, x2]

    return (x1-overscan, x2-overscan, y1, y2)


def trim_filter_edge(fname, x1, x2, y1, y2, output='', output_dir=''):
    """Trim image edges"""
    # Get median profile along slit:
    msg = list()
    msg.append("          - Loaded file: %s" % fname)

    if output == '':
        basename = os.path.basename(fname)
        output = 'trim_' + basename

    if output_dir != '':
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        output = os.path.join(output_dir, output)

    with fits.open(fname) as hdu_list:
        for hdu in hdu_list:
            if hdu.data is None:
                continue

            data = hdu.data
            hdr = hdu.header
            data_trim = data[y1:y2, x1:x2]
            hdr['CRPIX1'] -= x1
            hdr['NAXIS1'] = data_trim.shape[1]
            hdr['CRPIX2'] -= y1
            hdr['NAXIS2'] = data_trim.shape[0]
            hdr.add_comment("Image trimmed by PyNOT")
            hdu.data = data_trim
            hdu.header = hdr
        hdu_list.writeto(output, overwrite=True)
    msg.append(" [OUTPUT] - Saving trimmed image: %s" % output)
    msg.append("")
    output_msg = "\n".join(msg)

    return output_msg


def fit_background_image(data, order_bg=3, xmin=0, xmax=None, kappa=5, fwhm_scale=1):
    """
    Fit background in 2D spectral data. The background is fitted along the spatial rows by a Chebyshev polynomium.

    Parameters
    ==========
    data : np.array(M, N)
        Data array holding the input image

    order_bg : integer  [default=3]
        Order of the Chebyshev polynomium to fit the background

    xmin, xmax : integer  [default=0, None]
        Mask out pixels below xmin and above xmax

    fwhm_scale : float  [default=3]
        Number of FWHM below and above centroid of auto-detected trace
        that will be masked out during fitting.

    kappa : float  [default=10]
        Threshold for masking out cosmic rays etc.

    Returns
    =======
    bg2D : np.array(M, N)
        Background model of the 2D frame, same shape as input data.
    """
    x = np.arange(data.shape[1])
    if xmax is None:
        xmax = len(x)
    if xmax < 0:
        xmax = len(x) + xmax
    SPSF = np.nanmedian(data, 0)
    noise = 1.5*mad(SPSF)
    peaks, properties = signal.find_peaks(SPSF, prominence=10*noise, width=3)
    mask = (x >= xmin) & (x <= xmax)
    for num, center in enumerate(peaks):
        width = properties['widths'][num]
        x1 = center - width*fwhm_scale
        x2 = center + width*fwhm_scale
        obj = (x >= x1) * (x <= x2)
        mask &= ~obj

    bg2D = np.zeros_like(data)
    for i, row in enumerate(data):
        # Median filter the data to remove outliers:
        med_row = median_filter(row, 15)
        noise = mad(row)*1.4826
        this_mask = mask * (np.abs(row - med_row) < kappa*noise)
        if np.sum(this_mask) > order_bg+1:
            bg_model = Chebyshev.fit(x[this_mask], row[this_mask], order_bg, domain=[x.min(), x.max()])
            bg2D[i] = bg_model(x)

    return bg2D


def auto_fit_background(data_fname, output_fname, dispaxis=2, order_bg=3, kappa=10, fwhm_scale=3, xmin=0, xmax=None, plot_fname='', **kwargs):
    """
    Fit background in 2D spectral data. The background is fitted along the spatial rows by a Chebyshev polynomium.

    Parameters
    ==========
    data_fname : string
        Filename of the FITS image to process

    output_fname : string
        Filename of the output FITS image containing the background subtracted image
        as well as the background model in a separate extension.

    dispaxis : integer  [default=1]
        Dispersion axis, 1: horizontal spectra, 2: vertical spectra
        The function does not rotate the final image, only the intermediate image
        since it's faster to operate on rows than on columns.

    order_bg : integer  [default=3]
        Order of the Chebyshev polynomium to fit the background

    xmin, xmax : integer  [default=0, None]
        Mask out pixels below xmin and above xmax

    fwhm_scale : float  [default=3]
        Number of FWHM below and above centroid of auto-detected trace
        that will be masked out during fitting.

    kappa : float  [default=10]
        Threshold for masking out cosmic rays etc.

    plot_fname : string  [default='']
        Filename of diagnostic plots. If nothing is given, do not plot.

    kwargs : dict
        This has no effect. It's just to catch the unused options passed from the pipeline.

    Returns
    =======
    output_fname : string
        Background model of the 2D frame, same shape as input data.

    output_msg : string
        Log of messages from the function call
    """
    msg = list()
    data = fits.getdata(data_fname)
    hdr = get_header(data_fname)
    if 'DISPAXIS' in hdr:
        dispaxis = hdr['DISPAXIS']

    if dispaxis == 1:
        # transpose the horizontal spectra to make them vertical
        # since it's faster to fit rows than columns
        data = data.T
    msg.append("          - Loaded input image: %s" % data_fname)

    msg.append("          - Fitting background along the spatial axis with polynomium of order: %i" % order_bg)
    msg.append("          - Automatic masking of outlying pixels and object trace")
    bg2D = fit_background_image(data, order_bg=order_bg, kappa=kappa, fwhm_scale=fwhm_scale, xmin=xmin, xmax=xmax)

    if plot_fname:
        fig2D = plt.figure()
        ax1_2d = fig2D.add_subplot(121)
        ax2_2d = fig2D.add_subplot(122)
        noise = mad(data)
        v1 = np.median(data) - 5*noise
        v2 = np.median(data) + 5*noise
        ax1_2d.imshow(data, origin='lower', vmin=v1, vmax=v2, aspect='auto')
        ax1_2d.set_title("Input Image")
        ax2_2d.imshow(data-bg2D, origin='lower', vmin=v1, vmax=v2, aspect='auto')
        ax2_2d.set_title("Background Subtracted")
        ax1_2d.set_xlabel("Spatial Axis  [pixels]")
        ax2_2d.set_xlabel("Spatial Axis  [pixels]")
        ax1_2d.set_ylabel("Dispersion Axis  [pixels]")
        fig2D.tight_layout()
        fig2D.savefig(plot_fname)
        plt.close()
        msg.append(" [OUTPUT] - Saving diagnostic figure: %s" % plot_fname)

    data = data - bg2D
    if dispaxis == 1:
        # Rotate the model and data back to horizontal orientation:
        data = data.T
        bg2D = bg2D.T

    with fits.open(data_fname) as hdu:
        hdu[0].data = data
        sky_hdr = fits.Header()
        sky_hdr['BUNIT'] = 'count'
        copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1']
        copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2']
        sky_hdr['CTYPE2'] = 'LINEAR'
        sky_hdr['CUNIT2'] = 'Pixel'
        for key in copy_keywords:
            sky_hdr[key] = hdr[key]
        sky_hdr['AUTHOR'] = 'PyNOT version %s' % __version__
        sky_hdr['ORDER'] = (order_bg, "Polynomial order along spatial rows")
        sky_ext = fits.ImageHDU(bg2D, header=sky_hdr, name='SKY')
        hdu.append(sky_ext)
        hdu.writeto(output_fname, overwrite=True)

    msg.append(" [OUTPUT] - Saving background subtracted image: %s" % output_fname)
    msg.append("")
    output_msg = "\n".join(msg)
    return output_msg


def correct_cosmics(input_fname, output_fname, niter=4, gain=None, readnoise=None, sigclip=4.5, sigfrac=0.3, objlim=5.0, satlevel=113500.0, cleantype='meanmask'):
    """
    Detect and Correct Cosmic Ray Hits based on the method by van Dokkum (2001)
    The corrected frame is saved to a FITS file.

    Parameters
    ----------
    input_fname : string
        Input filename of 2D image

    output_fname : string
        Filename of corrected 2D image

    For details on other parameters, see `astroscrappy.detect_cosmics`

    Returns
    -------
    output_msg : string
        Log of messages from the function call
    """
    msg = list()
    msg.append("          - Cosmic Ray Rejection using Astroscrappy (based on van Dokkum 2001)")
    sci = fits.getdata(input_fname)
    hdr = fits.getheader(input_fname)
    hdr['EXTNAME'] = 'DATA'
    msg.append("          - Loaded input image: %s" % input_fname)
    with fits.open(input_fname) as hdu:
        if 'SKY' in hdu:
            sky_level = np.median(hdu['SKY'].data)
            msg.append("          - Image has been sky subtracted. Median sky level: %.1f" % sky_level)
        else:
            sky_level = 0.

        if 'MASK' in hdu:
            mask = hdu['MASK'].data
        else:
            mask = np.zeros_like(sci, dtype=int)

    if not gain:
        gain = hdr['GAIN']
        msg.append("          - Read GAIN from FITS header: %.2f" % gain)
    if not readnoise:
        try:
            readnoise = hdr['RDNOISE']
            msg.append("          - Read RDNOISE from FITS header: %.2f" % readnoise)
        except:
            readnoise = hdr['READNOISE']
            msg.append("          - Read READNOISE from FITS header: %.2f" % readnoise)

    crr_mask, sci = detect_cosmics(sci, gain=gain, readnoise=readnoise, niter=niter, pssl=sky_level,
                                   sigclip=sigclip, sigfrac=sigfrac, objlim=objlim, satlevel=satlevel,
                                   cleantype=cleantype)
    # Corrected image is in ELECTRONS!! Convert back to ADUs:
    sci = sci/gain - sky_level

    # Add comment to FITS header:
    hdr.add_comment("Cosmic Ray Rejection using Astroscrappy (based on van Dokkum 2001)")
    # expand mask to neighbouring pixels:
    msg.append("          - Number of cosmic ray hits identified: %i" % np.sum(mask > 0))

    mask = mask + 1*(crr_mask > 0)
    with fits.open(input_fname) as hdu:
        if hdu[0].data is not None:
            hdu[0].data = sci
            hdu[0].header = hdr
        else:
            hdu[1].data = sci
            hdu[1].header = hdr

        if 'MASK' in hdu:
            hdu['MASK'].data = mask
            hdu['MASK'].header.add_comment("Cosmic Ray Rejection using Astroscrappy (based on van Dokkum 2001)")
        else:
            mask_hdr = fits.Header()
            mask_hdr.add_comment("0 = Good Pixels")
            mask_hdr.add_comment("1 = Cosmic Ray Hits")
            mask_hdr.add_comment("Cosmic Ray Rejection using Astroscrappy (based on van Dokkum 2001)")
            mask_ext = fits.ImageHDU(mask, header=mask_hdr, name='MASK')
            hdu.append(mask_ext)
        hdu.writeto(output_fname, overwrite=True)
    msg.append(" [OUTPUT] - Saving cosmic ray corrected image: %s" % output_fname)
    msg.append("")
    output_msg = "\n".join(msg)
    return output_msg


def raw_correction(sci_raw, hdr, bias_fname, flat_fname='', output='', overscan=50, overwrite=True, mode='spec'):
    """
    Perform bias subtraction, flat field correction, and cosmic ray rejection

    Parameters
    ==========

    sci_raw : np.array (M, N)
        Input science image to reduce

    hdr : FITS Header
        Header associated with the science image

    bias_fname : string
        Filename of bias image to subtract from `sci_raw`

    flat_fname : string  [default='']
        Filename of normalized flat field image. If none is given, no flat field correction will be applied

    output : string  [default='']
        Output filename

    overscan : int  [default=50]
        Number of pixels in overscan at the edge of the CCD.
        The overscan region will be trimmed.

    overwrite : boolean  [default=True]
        Overwrite existing output file if True.

    Returns
    -------
    output_msg : string
        Log of status messages
    """
    msg = list()
    mbias = fits.getdata(bias_fname)
    # bias_hdr = get_header(bias_fname)
    msg.append("          - Loaded bias image: %s" % bias_fname)
    if flat_fname:
        mflat = fits.getdata(flat_fname)
        mflat[mflat == 0] = 1
        # flat_hdr = get_header(flat_fname)
        msg.append("          - Loaded flat field image: %s" % flat_fname)
        # mflat, flat_hdr = trim_overscan(mflat, flat_hdr, overscan=overscan, mode=mode)
    else:
        mflat = 1.
        msg.append("          - Not flat field image provided. No correction applied!")

    # Trim overscan of raw image:
    # - Trimming again of processed images doesn't change anything,
    # - so do it just in case the input has not been trimmed
    sci_raw, hdr = trim_overscan(sci_raw, hdr, overscan=overscan, mode=mode)
    # mbias, bias_hdr = trim_overscan(mbias, bias_hdr, overscan=overscan, mode=mode)

    # Correct image:
    sci = (sci_raw - mbias)/mflat

    # Calculate error image:
    if hdr['CCDNAME'] == 'CCD14':
        hdr['GAIN'] = 0.16
    gain = hdr['GAIN']
    readnoise = hdr['RDNOISE']
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        err = np.sqrt(gain*sci + readnoise**2) / gain
    msg.append("          - Created noise image")
    msg.append("          - Gain=%.2f  and Read Noise=%.2f" % (gain, readnoise))

    # Fix NaN values from negative pixel values:
    err_NaN = np.isnan(err)
    err[err_NaN] = readnoise/gain
    msg.append("          - Correcting NaNs in noise image: %i pixel(s)" % np.sum(err_NaN))
    hdr['DATAMIN'] = np.nanmin(sci)
    hdr['DATAMAX'] = np.nanmax(sci)
    hdr['EXTNAME'] = 'DATA'
    hdr['AUTHOR'] = 'PyNOT version %s' % __version__
    hdr['OVERSCAN'] = 'TRIMMED'

    mask = np.zeros_like(sci, dtype=int)
    msg.append("          - Empty pixel mask created")
    mask_hdr = fits.Header()
    mask_hdr.add_comment("0 = Good Pixels")
    mask_hdr.add_comment("1 = Cosmic Ray Hits")
    for key in ['CRPIX1', 'CRPIX2', 'CRVAL1', 'CRVAL2', 'CTYPE1', 'CTYPE2', 'CUNIT1', 'CUNIT2']:
        mask_hdr[key] = hdr[key]

    if mode == 'spec':
        mask_hdr['CDELT1'] = hdr['CDELT1']
        mask_hdr['CDELT2'] = hdr['CDELT2']
    else:
        mask_hdr['CD1_1'] = hdr['CD1_1']
        mask_hdr['CD1_2'] = hdr['CD1_2']
        mask_hdr['CD2_1'] = hdr['CD2_1']
        mask_hdr['CD2_2'] = hdr['CD2_2']

    sci_ext = fits.PrimaryHDU(sci, header=hdr)
    err_ext = fits.ImageHDU(err, header=hdr, name='ERR')
    mask_ext = fits.ImageHDU(mask, header=mask_hdr, name='MASK')
    output_HDU = fits.HDUList([sci_ext, err_ext, mask_ext])
    output_HDU.writeto(output, overwrite=overwrite)
    msg.append("          - Successfully corrected the image.")
    msg.append(" [OUTPUT] - Saving output: %s" % output)
    msg.append("")
    output_msg = "\n".join(msg)

    return output_msg



def correct_raw_file(input_fname, *, output, bias_fname, flat_fname='', overscan=50, overwrite=True, mode='spec'):
    """
    Wrapper for `raw_correction` using file input instead of image input

    Returns
    -------
    output_msg : string
        Log of status messages
    """
    hdr = get_header(input_fname)
    sci_raw = fits.getdata(input_fname)
    msg = "          - Loaded input image: %s" % input_fname

    output_msg = raw_correction(sci_raw, hdr, bias_fname, flat_fname, output=output,
                                overscan=overscan, overwrite=overwrite, mode=mode)
    output_msg = msg + '\n' + output_msg

    return output_msg



# -*- coding: UTF-8 -*-

import numpy as np
import os
import sys
import datetime
from glob import glob
from astropy.io import fits

from pynot.alfosc import get_binning_from_hdr, get_filter, get_header
from pynot.response import lookup_std_star
import pynot.alfosc as instrument

# -- use os.path
code_dir = os.path.dirname(os.path.abspath(__file__))
v_file = os.path.join(os.path.split(code_dir)[0], 'VERSION')
with open(v_file) as version_file:
    __version__ = version_file.read().strip()

std_fname = os.path.join(os.path.split(code_dir)[0], 'calib/std/namelist.txt')
calib_namelist = np.loadtxt(std_fname, dtype=str)
calib_names = calib_namelist[:, 1]

alfosc_rulefile = os.path.join(code_dir, 'alfosc.rules')

def occurence(inlist):
    """
    Counts the number of occurences of each number in the input list.

    Returns
    =======
    single_values : list
        List of tuples (v, N(v)), where v is any value in `inlist`
        and N(v) is the number of occurences of this value.
    """
    inlist = np.array(inlist)
    n = np.zeros(len(inlist), dtype=bool)
    single_values = list()
    while np.sum(n) < len(inlist):
        e0 = inlist[~n][0]
        Ne = np.sum(inlist == e0)
        n += inlist == e0
        single_values.append((e0, Ne))
    return single_values


def get_mjd(date_str):
    """Input Date String in ISO format as in ALFOSC header: '2016-08-01T16:03:57.796'"""
    date = datetime.datetime.fromisoformat(date_str)
    mjd_0 = datetime.datetime(1858, 11, 17)
    dt = date - mjd_0
    mjd = dt.days + dt.seconds/(24*3600.)
    return mjd


def match_date(files, date_mjd):
    """
    Return subset of files which were observed at the given date.
    Input
    =====

    files : list(str)
        List of filenames to be matched to the given date

    date_mjd : float
        Modified Julian date to match, MJD
    """
    matches = list()
    for fname in files:
        hdr = fits.getheader(fname)
        mjd = get_mjd(hdr['DATE-OBS'])
        if int(mjd) == int(date_mjd):
            matches.append(fname)

    return matches


def group_calibs_by_date(file_list, lower=0.01, upper=0.99):
    """ Grouping files observed from the same day. """
    file_list = np.array(file_list, dtype=str)
    date_sorted = dict()
    mjd = np.zeros(len(file_list))
    n = np.zeros(len(file_list), dtype=bool)
    N = len(file_list)
    for i, fname in enumerate(file_list):
        hdr = fits.getheader(fname)
        mjd[i] = get_mjd(hdr['DATE-OBS'])

    while sum(n) < N:
        m0 = mjd[~n][0]
        this_date = (mjd - np.floor(m0) > lower) * (mjd - np.floor(m0) < upper)
        date_string = str(int(m0) - 1)
        date_sorted[date_string] = list(file_list[this_date])
        n = n + this_date

    return date_sorted


def parse_value(val):
    val = val.strip()
    if '.' in val:
        if val.replace('.', '').isnumeric():
            new_val = float(val)
    elif val.isnumeric():
        new_val = int(val)
    else:
        new_val = val
    return new_val


def get_unclassified_files(file_list, database):
    missing_files = list()
    for fname in file_list:
        if fname not in database.file_database:
            missing_files.append(fname)
    return missing_files


def classify_file(fname, rules):
    """
    Classify input FITS file according to the set of `rules`, a list of string conditions
    for header keywords. Each rule correspond to one filetype, e.g. BIAS, OBJECT
    Returns a list of rules that
    """
    msg = list()
    try:
        h = fits.getheader(fname)
    except OSError:
        msg.append("File could not be opened: %s" % fname)
        return [], msg

    fileroot = fname.split('/')[-1]
    if fileroot != h['FILENAME']:
        msg.append("Filename does not match FILENAME in header: %s" % fname)
        return [], msg

    matches = list()
    for linenum, rule in enumerate(rules):
        rule = rule.strip()
        if ('#' in rule) or (len(rule) == 0):
            continue

        ftype = rule.split(':')[0].strip()
        all_conditions = rule.split(':')[1].split(' and ')
        criteria = list()
        for cond in all_conditions:
            if '==' in cond:
                key, val = cond.split('==')
                if key not in h:
                    return [], msg
                if 'open' in val.lower():
                    criteria.append('open' in h[key].lower())
                elif 'closed' in val.lower():
                    criteria.append('closed' in h[key].lower())
                else:
                    val = parse_value(val)
                    criteria.append(val == h[key])

            elif '!=' in cond:
                key, val = cond.split('!=')
                if key not in h:
                    return [], msg
                if 'open' in val.lower():
                    criteria.append('open' not in h[key].lower())
                elif 'closed' in val.lower():
                    criteria.append('closed' not in h[key].lower())
                else:
                    val = parse_value(val)
                    criteria.append(val != h[key])

            elif '>' in cond:
                key, val = cond.split('>')
                if key not in h:
                    return [], msg
                val = parse_value(val)
                criteria.append(h[key] > val)

            elif '<' in cond:
                key, val = cond.split('<')
                if key not in h:
                    return [], msg
                val = parse_value(val)
                criteria.append(h[key] < val)

            elif ' contains ' in cond:
                key, val = cond.split('contains')
                if key not in h:
                    return [], msg
                val = parse_value(val)
                criteria.append(val in h[key])

            elif ' !contains ' in cond:
                key, val = cond.split('!contains')
                if key not in h:
                    return [], msg
                val = parse_value(val)
                criteria.append(val not in h[key])

            else:
                raise ValueError("Invalid condition in rule at line %i:  %s" % (linenum, rule))

        if np.all(criteria):
            matches.append(ftype)

    if len(matches) == 1:
        ftype = matches[0]
        if ftype == 'IMG_OBJECT':
            # Check if target is a flux standard calibrator!
            pass

        elif ftype == 'SPEC_OBJECT':
            if instrument.target_keyword in h:
                star_target = h[instrument.target_keyword]
                star_name = lookup_std_star(star_target)
                if star_name:
                    matches = ['SPEC_FLUX-STD']

    elif len(matches) == 0:
        msg.append("No classification matched the file: %s" % fname)
    else:
        msg.append("More than one classification was found for file: %s" % fname)
        msg.append(", ".join(matches))

    return matches, msg


def classify(data_in, rule_file=alfosc_rulefile, progress=True):
    """
    The input can be a single .fits file, a string given the path to a directory,
    a list of .fits files, or a list of directories.
    Classify given input 'files' using the rules defined in 'alfosc.rules'.
    Returns
    'Data_Type' containing the classification for the pipeline
                recipes.
    """
    msg = list()

    # Determine the input type:
    if isinstance(data_in, str):
        if os.path.isfile(data_in):
            files = [data_in]
        elif os.path.isdir(data_in):
            files = glob(data_in+'/*.fits')

    elif isinstance(data_in, list):
        if os.path.isfile(data_in[0]):
            files = data_in
        elif os.path.isdir(data_in[0]):
            files = []
            for path in data_in:
                files += glob(path+'/*.fits')
    else:
        raise ValueError("Input must be a string or a list of strings")

    data_types = dict()
    not_classified_files = list()

    if not os.path.exists(rule_file):
        raise FileNotFoundError("ALFOSC ruleset could not be found: %s" % rule_file)

    with open(rule_file) as rulebook:
        rules = rulebook.readlines()

    if progress:
        print("")
        print(" Classifying files: ")

    for num, fname in enumerate(files):
        matches, output_msg = classify_file(fname, rules)

        if len(matches) == 1:
            ftype = matches[0]
            data_types[fname] = ftype

        elif len(matches) == 0:
            msg.append(" [ERROR]  - " + output_msg[0])
            not_classified_files.append(fname)

        else:
            msg.append(" [ERROR]  - " + output_msg[0])
            msg.append("            " + output_msg[1])

        if progress:
            sys.stdout.write("\r  %6.2f%%" % (100.*(num+1)/len(files)))
            sys.stdout.flush()

    msg.append("")
    msg.append("          - Classification finished.")
    msg.append("          - Successfully classified %i out of %i files." % (len(data_types.keys()), len(files)))
    msg.append("")
    if len(files) != len(data_types.keys()):
        msg.append("[WARNING] - Files not classified:")
        for item in not_classified_files:
            msg.append("          - %s" % item)
    msg.append("")
    output_msg = "\n".join(msg)

    database = TagDatabase(data_types)

    return database, output_msg


def reclassify(data_in, database, **kwargs):
    # Determine the input type:
    if isinstance(data_in, str):
        if os.path.isfile(data_in):
            files = [data_in]
        elif os.path.isdir(data_in):
            files = glob(data_in+'/*.fits')

    elif isinstance(data_in, list):
        if os.path.isfile(data_in[0]):
            files = data_in
        elif os.path.isdir(data_in[0]):
            files = []
            for path in data_in:
                files += glob(path+'/*.fits')
    else:
        raise ValueError("Input must be a string or a list of strings")

    missing_files = get_unclassified_files(files, database)

    if len(missing_files) == 0:
        output_msg = "No new files to classify!"
        new_database = database
    else:
        new_database, output_msg = classify(missing_files, **kwargs)
        new_database = database + new_database

    return new_database, output_msg



def write_report(collection, output=''):
    """
    Write a report of a given data collection from the `classify` function

    If `output` is given, the report will be saved to this filename.
    """
    # Sort the files by file-type:
    tags = TagDatabase(collection)

    report = "# Data Report for ALFOSC\n"
    for tag in sorted(tags.keys()):
        report += "\n - %s:\n" % tag
        for filename in sorted(tags[tag]):
            report += " %s\n" % filename
    if output == '':
        print(report)
    else:
        with open(output, 'w') as output:
            output.write(report)


class RawImage(object):
    def __init__(self, fname, filetype=None):
        self.filename = fname
        self.data = fits.getdata(fname)
        self.shape = self.data.shape
        self.filetype = filetype
        # Merge Primary and Image headers:
        self.header = get_header(fname)
        # primhdr = fits.getheader(fname, 0)
        # imghdr = fits.getheader(fname, 1)
        # primhdr.update(imghdr)
        # self.header = primhdr
        self.binning = get_binning_from_hdr(self.header)
        file_root = fname.split('/')[-1]
        self.dispaxis = None
        if file_root != self.header['FILENAME']:
            raise ValueError("The file doesn't seem to be a raw FITS image.")

        if self.header['OBS_MODE'] == 'SPECTROSCOPY':
            self.mode = 'SPEC'
            cd1 = self.header['CDELT1']
            cd2 = self.header['CDELT2']
            self.CD = np.array([[cd1, 0.],
                                [0., cd2]])
        else:
            self.mode = 'IMG'
            # Image Coordinates and Reference System
            cd11 = self.header['CD1_1']
            cd12 = self.header['CD1_2']
            cd21 = self.header['CD2_1']
            cd22 = self.header['CD2_2']
            self.CD = np.array([[cd11, cd21],
                                [cd12, cd22]])

        self.filter = get_filter(self.header)
        self.slit = self.header['ALAPRTNM']
        self.grism = self.header['ALGRNM']
        if 'Vert' in self.slit:
            self.dispaxis = 1
        elif 'Slit' in self.slit:
            self.dispaxis = 2

        self.exptime = self.header['EXPTIME']
        self.object = self.header['OBJECT']
        self.target_name = self.header['TCSTGT']
        self.ra_deg = self.header['RA']
        self.ra_hr = self.header['OBJRA']
        self.dec_deg = self.header['DEC']

        self.rot_angle = self.header['ROTPOS']
        self.airmass = self.header['AIRMASS']
        self.date = self.header['DATE-OBS']
        self.mjd = get_mjd(self.date)

        self.CRVAL = np.array([self.header['CRVAL1'], self.header['CRVAL2']])
        self.CRPIX = np.array([self.header['CRPIX1'], self.header['CRPIX2']])
        self.data_unit = self.header['BUNIT']
        self.x_unit = self.header['CUNIT1']
        self.y_unit = self.header['CUNIT2']

    def match_files(self, filelist, date=True, binning=True, shape=True, grism=False, slit=False, filter=False, get_closest_time=False):
        """Return list of filenames that match the given criteria"""
        matches = list()
        # sort by:
        all_times = list()
        for fname in filelist:
            this_hdr = fits.getheader(fname, 0)
            criteria = list()
            this_mjd = get_mjd(this_hdr['DATE-OBS'])
            if date:
                # Match files from same night, midnight ± 9hr
                dt = self.mjd - this_mjd
                criteria.append(-0.4 < dt < +0.4)

            if binning:
                # Match files with same binning and readout speed:
                this_binning = get_binning_from_hdr(this_hdr)
                criteria.append(this_binning == self.binning)

            if shape:
                # Match files with same image shape:
                this_shape = fits.getdata(fname).shape
                criteria.append(this_shape == self.shape)

            if grism:
                # Match files with same grism:
                this_grism = this_hdr['ALGRNM']
                criteria.append(this_grism == self.grism)

            if slit:
                # Match files with the same slit-width:
                this_slit = this_hdr['ALAPRTNM']
                criteria.append(this_slit == self.slit)

            if filter:
                # Match files with the same filter:
                this_filter = get_filter(this_hdr)
                criteria.append(this_filter == self.filter)

            if np.all(criteria):
                matches.append(fname)
                all_times.append(this_mjd)

        if get_closest_time:
            index = np.argmin(np.abs(np.array(all_times) - self.mjd))
            matches = matches[index:index+1]

        return matches



class TagDatabase(dict):
    def __init__(self, file_database):
        # Convert file_database with file-classifications
        # to a tag_database containing a list of all files with a given tag:
        tag_database = dict()
        for fname, tag in file_database.items():
            if tag in tag_database.keys():
                tag_database[tag].append(fname)
            else:
                tag_database[tag] = [fname]

        # And make this converted tag_database the basis of the TagDatabase class
        dict.__init__(self, tag_database)
        self.file_database = file_database

    def __add__(self, other):
        new_file_database = other.file_database

        for key in self.file_database.keys():
            new_file_database[key] = self.file_database[key]

        return TagDatabase(new_file_database)

    def __radd__(self, other):
        if other == 0:
            return self

        else:
            return self.__add__(other)

    def has_tag(self, tag):
        if tag in self.keys():
            return True

        return False



"""
Instrument definitions for NOT/ALFOSC
"""
import numpy as np
from os.path import dirname, abspath
from astropy.io import fits
from astropy.table import Table


# Define header keyword to use as object name
# OBJECT is not always reliable, TCS target name is more robust
target_keyword = 'TCSTGT'

# path = '/Users/krogager/coding/PyNOT'
path = dirname(abspath(__file__))

# path to extinction table:
extinction_fname = path + '/calib/lapalma.ext'

grism_translate = {'Grism_#3': 'grism3',
                   'Grism_#4': 'grism4',
                   'Grism_#5': 'grism5',
                   'Grism_#6': 'grism6',
                   'Grism_#7': 'grism7',
                   'Grism_#8': 'grism8',
                   'Grism_#10': 'grism10',
                   'Grism_#11': 'grism11',
                   'Grism_#12': 'grism12',
                   'Grism_#14': 'grism14',
                   'Grism_#15': 'grism15',
                   'Grism_#16': 'grism16',
                   'Grism_#17': 'grism17',
                   'Grism_#18': 'grism18',
                   'Grism_#19': 'grism19',
                   'Grism_#20': 'grism20'}


slits = ['Ech_0.7', 'Ech_0.8', 'Ech_1.0', 'Ech_1.2', 'Ech_1.6',
         'Ech_1.8', 'Ech_2.2', 'Pol_1.0', 'Pol_1.4', 'Pol_1.8',
         'Slit_0.4', 'Slit_0.5', 'Slit_0.75', 'Slit_1.0',
         'Slit_1.3', 'Slit_1.8', 'Slit_2.5', 'Slit_10.0',
         'Vert_0.5', 'Vert_0.75', 'Vert_0.9', 'Vert_1.3',
         'Vert_1.8', 'Vert_10.0', 'VertOff_0.5', 'VertOff_0.8',
         'VertOff_1.0', 'VertOff_1.3', 'VertOff_1.9', 'VertOff_8.8']


filter_table = Table.read(path + '/calib/alfosc_filters.dat', format='ascii.fixed_width')
filter_translate = {long: short for long, short in filter_table['name', 'short_name']}


def get_header(fname):
    with fits.open(fname) as hdu:
        primhdr = hdu[0].header
        if len(hdu) > 1:
            imghdr = hdu[1].header
            primhdr.update(imghdr)
    if primhdr['INSTRUME'] != 'ALFOSC_FASU':
        print("[WARNING] - FITS file not originating from NOT/ALFOSC!")
    return primhdr


def create_pixel_array(hdr, dispaxis):
    """Load reference array from header using CRVAL, CDELT, CRPIX along dispersion axis"""
    if dispaxis not in [1, 2]:
        raise ValueError("Dispersion Axis must be 1 (X-axis) or 2 (Y-axis)!")
    p = hdr['CRVAL%i' % dispaxis]
    s = hdr['CDELT%i' % dispaxis]
    r = hdr['CRPIX%i' % dispaxis]
    N = hdr['NAXIS%i' % dispaxis]
    # -- If data are from NOT then check for binning and rescale CRPIX:
    binning = 1
    if 'DETXBIN' in hdr:
        if dispaxis == 1:
            binning = hdr['DETXBIN']
        else:
            binning = hdr['DETYBIN']
    pix_array = p + s*(np.arange(N) - (r/binning - 1))
    return pix_array


def get_binning(fname):
    hdr = fits.getheader(fname)
    binx = hdr['DETXBIN']
    biny = hdr['DETYBIN']
    read = hdr['FPIX']
    ccd_setup = "%ix%i_%i" % (binx, biny, read)
    return ccd_setup


def get_binning_from_hdr(hdr):
    binx = hdr['DETXBIN']
    biny = hdr['DETYBIN']
    read = hdr['FPIX']
    ccd_setup = "%ix%i_%i" % (binx, biny, read)
    return ccd_setup


def get_filter(hdr):
    filter = 'Open'
    for keyword in ['FAFLTNM', 'FBFLTNM', 'ALFLTNM']:
        if 'open' in hdr[keyword].lower():
            pass
        else:
            filter = hdr[keyword]
            break
    if '  ' in filter:
        filter = filter.replace('  ', ' ')
    return filter

def get_grism(hdr):
    grism_name = grism_translate(hdr['ALGRNM'])
    return grism_name

def get_slit(hdr):
    return hdr['ALAPRTNM']

def get_airmass(hdr):
    """Return the average airmass at mid-exposure"""
    return hdr['AIRMASS']



from astropy.wcs import WCS
from astropy.io import fits
from astropy.table import Table
from astropy.utils.exceptions import AstropyWarning

import numpy as np
import matplotlib.pyplot as plt
import os
import warnings

from pynot.wcs import get_gaia_catalog
from pynot.functions import decimal_to_string
from pynot import alfosc


def mad(x):
    return np.median(np.abs(x-np.median(x)))


def find_sources_without_gaia(sep_cat, gaia, limit=1.5):
    no_match_list = list()
    refs = np.array([gaia['ra'], gaia['dec']]).T
    for row in sep_cat:
        xy = np.array([row['ra'], row['dec']])
        dist = np.sqrt(np.sum((refs - xy)**2, axis=1))
        if np.min(dist) < limit/3600.:
            pass
        else:
            no_match_list.append(np.array(row))
    no_match_list = np.array(no_match_list)
    return Table(no_match_list)


def find_new_sources(img_fname, sep_fname, loc_bat=(0., 0., 1.), loc_xrt=(0., 0., 1), mag_lim=20.1, zp=None):
    """
    Cross-match image source catalog with Gaia source catalog to identify new sources.

    Parameters
    ----------
    img_fname : string
        Filename of WCS calibrated image (_wcs.fits)

    sep_fname : string
        Filename of the source extraction table (_phot.fits)

    loc_bat : tuple(ra, dec, radius)
        Localisation of gamma ray detection, ra, dec and radius in deg

    loc_xrt : tuple(ra, dec, radius)
        Localisation of X-ray detection, ra, dec and radius in deg

    mag_lim : float
        Magnitude limit in the given filter

    zp : float  [default=None]
        Magnitude zero point if the source catalog has not been flux calibrated

    Returns
    -------
    new_subset : astropy.table.Table
        Subset of the source catalog which are not in the Gaia catalog above the flux limit.
        Contains the following columns: 'ra', 'dec', 'mag_auto', 'a', 'b', 'theta', 'flux_auto', 'flux_err_auto'

    output_msg : string
        Log of messages from the function call.
    """
    msg = list()
    img = fits.getdata(img_fname)
    hdr = fits.getheader(img_fname)

    # Download Gaia positions:
    base, ext = os.path.splitext(os.path.basename(img_fname))
    dirname = os.path.dirname(img_fname)
    image_radius = np.sqrt(hdr['NAXIS1']**2 + hdr['NAXIS2']**2) / 2
    image_scale = np.sqrt(hdr['CD1_1']**2 + hdr['CD1_2']**2)
    deg_to_arcmin = 60.
    radius = image_scale * image_radius * deg_to_arcmin
    gaia_cat_name = 'gaia_source_%.2f%+.2f_%.1f.csv' % (hdr['CRVAL1'], hdr['CRVAL2'], radius)
    gaia_cat_name = os.path.join(dirname, gaia_cat_name)
    gaia_dr = 'edr3'
    if os.path.exists(gaia_cat_name):
        msg.append("          - Loading Gaia source catalog: %s" % gaia_cat_name)
        ref_cat = Table.read(gaia_cat_name)
    else:
        msg.append("          - Downloading Gaia source catalog... (%s)" % gaia_dr.upper())
        ref_cat = get_gaia_catalog(hdr['CRVAL1'], hdr['CRVAL2'], radius=radius,
                                   catalog_fname=gaia_cat_name, database=gaia_dr)


    # Get the source extraction catalog:
    sep_cat = Table.read(sep_fname)
    sep_hdr = fits.getheader(sep_fname)
    if 'MAG_ZP' not in sep_hdr:
        if zp is None:
            msg.append("[WARNING] - The source table has not been flux calibrated yet. Run task:")
            msg.append("                ]%% pynot autozp %s %s" % (img_fname, sep_fname))
            msg.append("            or provide a zeropoint using the option: -z")
            msg.append("          - Terminating script...")
            msg.append("")
            return [], "\n".join(msg)
        else:
            try:
                sep_cat['mag_auto'] += zp
            except ValueError:
                msg.append(" [ERROR]  - Invalid zeropoint: %r" % zp)
                msg.append("")
                return [], "\n".join(msg)

    # Find sources with r < 20, with no match in Gaia:
    filter = alfosc.filter_translate[alfosc.get_filter(hdr)]
    if '_' in filter:
        band = filter.split('_')[0]

    mag_band = sep_cat['mag_auto']
    new_sources = find_sources_without_gaia(sep_cat[mag_band < mag_lim], ref_cat)
    warnings.simplefilter('ignore', category=AstropyWarning)
    wcs = WCS(hdr)
    if len(new_sources) > 0:
        mag = new_sources['mag_auto']
        q_ab = new_sources['b']/new_sources['a']
        new = new_sources[(q_ab > 0.80) & (mag > 12)]

        fig = plt.figure(figsize=(9, 6))
        ax = fig.add_subplot(111, projection=wcs)
        med_val = np.nanmedian(img)
        ax.imshow(img, vmin=med_val-1*mad(img), vmax=med_val+10*mad(img),
                  origin='lower', cmap=plt.cm.gray_r)
        ylims = ax.get_ylim()
        xlims = ax.get_xlim()
        ax.scatter(ref_cat['ra'], ref_cat['dec'], label='Gaia',
                   transform=ax.get_transform('fk5'),
                   edgecolor='b', facecolor='none', s=40, marker='s')
        ax.set_xlabel("Right Ascension")
        ax.set_ylabel("Declination")
        box = ax.get_position()
        ax.set_position([0.15, box.y0, box.width, box.height])

        sattelite_data = (loc_bat, loc_xrt)
        sat_names = ('BAT', 'XRT')
        linestyles = (':', '--')
        colors = ('k', 'red')
        linewidths = (1., 1.0)
        chi2 = np.zeros_like(new['ra'])
        inside_bat = np.zeros(len(new), dtype=bool)
        for sat_name, burst_data, ls, col, lw in zip(sat_names, sattelite_data, linestyles, colors, linewidths):
            alpha, dec, sigma = burst_data
            err_circle = plt.Circle((alpha, dec), radius=sigma,
                                    transform=ax.get_transform('fk5'),
                                    facecolor='none', edgecolor=col, ls=ls, lw=lw,
                                    label=sat_name)
            ax.add_patch(err_circle)
            # convert 90% CI to 1-sigma (Gaussian):
            sig1 = sigma / 1.66
            # Calculate deviation in sigmas:
            delta_r = np.sqrt((new['ra']-alpha)**2 + (new['dec']-dec)**2)
            if sig1 == 0:
                pass
            else:
                chi2 += delta_r**2 / sig1**2
                if sat_name == 'BAT':
                    inside_bat = delta_r / sigma <= 1
        significance = np.sqrt(chi2)

        msg.append("          - New sources with no Gaia match")
        msg.append("          ---------------------------------")
        for source_id, (row, sig, in_bat) in enumerate(zip(new, significance, inside_bat)):
            if (sig < 5) and (sig > 0):
                color = 'Green'
                term_color = '\033[32m'      # green
                mark_good = '    [NEW] -'
            elif in_bat:
                color = 'DarkOrange'
                term_color = '\033[33m'      # yellow
                mark_good = 11*' '
            else:
                color = 'Red'
                term_color = '\033[31m'      # red
                mark_good = 11*' '

            mag = row['mag_auto']
            ax.scatter(row['ra'], row['dec'],
                       label='(%i)  %s = %.1f mag' % (source_id+1, band, mag),
                       transform=ax.get_transform('fk5'), edgecolor=color,
                       facecolor='none', s=60, lw=1.5)
            ax.text(row['ra']-8/3600., row['dec'], "%i" % (source_id+1),
                    transform=ax.get_transform('fk5'))

            reset = '\033[0m'
            ra_str, dec_str = decimal_to_string(row['ra'], row['dec'])
            msg.append(term_color + "%s %s %s  %s=%5.2f" % (mark_good, ra_str, dec_str, band, mag) + reset)
        ax.legend(title=hdr['OBJECT'].upper(), loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_ylim(*ylims)
        ax.set_xlim(*xlims)
        # Make a classification array for the type of new sources:
        # 0: not consistent with BAT nor XRT
        # 1: consistent with BAT
        # 2: consistent with BAT and XRT
        class_new = 1*inside_bat + 1*(significance < 5)*(significance > 0)
        new_subset = new['ra', 'dec', 'mag_auto', 'a', 'b', 'theta', 'flux_auto', 'flux_err_auto']
        new_subset['class'] = class_new
        new_table_fname = "new_sources_%s.txt" % base
        new_table_fname = os.path.join(dirname, new_table_fname)
        with open(new_table_fname, 'w') as new_table:
            units = ['(deg)', '(deg)', '(AB)', '(pix)', '(pix)', '(rad)', '(count/s)', '(count/s)', '']
            header = "{:^9}  {:^9}  {:^8}  {:^5}  {:^5}  {:^5}  {:^9}  {:^13}  {:^5}\n".format(*new_subset.colnames)
            unit_header = "{:^9}  {:^9}  {:^8}  {:^5}  {:^5}  {:^5}  {:^9}  {:^13}  {:^5}\n".format(*units)
            new_table.write(header)
            new_table.write(unit_header)
            np.savetxt(new_table, new_subset, fmt="%9.5f  %+9.5f  %8.2f  %5.1f  %5.1f  % 5.2f  %9.2e  %13.2e  %5i")
        msg.append(" [OUTPUT] - Writing detection table: %s" % new_table_fname)

        fig_fname = "new_sources_%s.pdf" % base
        fig_fname = os.path.join(dirname, fig_fname)
        fig.savefig(fig_fname)
        msg.append(" [OUTPUT] - Saving figure: %s" % fig_fname)

    else:
        new_subset = []
        msg.append("          - No new sources identified brighter than %s < %.2f" % (band, mag_lim))
    msg.append("")
    output_msg = "\n".join(msg)
    return new_subset, output_msg



# -*- coding: UTF-8 -*-

"""
Input / Output functions for the DataSet class.
"""

from astropy.io import fits
import numpy as np

from pynot.data.organizer import TagDatabase, get_filter

veclen = np.vectorize(len)


def get_header_info(fname):
    primhdr = fits.getheader(fname)
    imhdr = fits.getheader(fname, 1)
    if primhdr['INSTRUME'] != 'ALFOSC_FASU':
        raise ValueError("[WARNING] - FITS file not originating from NOT/ALFOSC!")
    object = primhdr['OBJECT']
    exptime = "%.1f" % primhdr['EXPTIME']
    grism = primhdr['ALGRNM']
    slit = primhdr['ALAPRTNM']
    filter = get_filter(primhdr)
    shape = "%ix%i" % (imhdr['NAXIS1'], imhdr['NAXIS2'])
    return object, exptime, grism, slit, filter, shape


def save_database(database, output_fname):
    """Save file database to file."""
    with open(output_fname, 'w') as output:
        output.write("# PyNOT File Classification Table\n\n")
        for filetype, files in sorted(database.items()):
            output.write("# %s:\n" % filetype)
            file_list = list()
            for fname in sorted(files):
                object, exptime, grism, slit, filter, shape = get_header_info(fname)
                file_list.append((fname, filetype, object, exptime, grism, slit, filter, shape))
            file_list = np.array(file_list, dtype=str)
            header_names = ('FILENAME', 'TYPE', 'OBJECT', 'EXPTIME', 'GRISM', 'SLIT', 'FILTER', 'SHAPE')
            max_len = np.max(veclen(file_list), 0)
            max_len = np.max([max_len, [len(n) for n in header_names]], 0)
            line_fmt = "  ".join(["%-{}s".format(n) for n in max_len])
            header = line_fmt % header_names
            output.write('#' + header + '\n')
            for line in file_list:
                output.write(' ' + line_fmt % tuple(line) + '\n')
            output.write("\n")


def load_database(input_fname):
    """Load file database from file."""
    all_lines = np.loadtxt(input_fname, dtype=str, usecols=(0, 1))
    file_database = {key: val for key, val in all_lines}
    return TagDatabase(file_database)




"""
Instrument definitions for NTT/EFOSC2
"""

import numpy as np
from os.path import dirname, abspath
from astropy.io import fits
from astropy.table import Table


# Define header keyword to use as object name
# OBJECT is not always reliable, TCS target name is more robust
target_keyword = 'OBJECT'
# or
# target_keyword = 'ESO OBS TARG NAME'

# path = '/Users/krogager/coding/PyNOT'
path = dirname(abspath(__file__))

# path to extinction table:
extinction_fname = path + '/calib/lasilla.ext'

# List grisms, slits and filters, and define filter translations if needed.

grism_translate = {'Gr#3': 'grism3',
                   'Gr#4': 'grism4',
                   'Gr#5': 'grism5',
                   'Gr#6': 'grism6',
                   'Gr#7': 'grism7',
                   'Gr#8': 'grism8',
                   'Gr#10': 'grism10',
                   'Gr#11': 'grism11',
                   'Gr#12': 'grism12',
                   'Gr#14': 'grism14',
                   'Gr#15': 'grism15',
                   'Gr#16': 'grism16',
                   'Gr#17': 'grism17',
                   'Gr#18': 'grism18',
                   'Gr#19': 'grism19',
                   'Gr#20': 'grism20'}

slit_translate = {'slit#1.0': 'slit_1.0',
                  'slit#1.2': 'slit_1.2',
                  'slit#1.5': 'slit_1.5',
                  'slit#2.0': 'slit_2.0',
                  'slit#0.7': 'slit_0.7',
                  'slit#0.5': 'slit_0.5',
                  'slit#5.0': 'slit_5.0',
                  }



def get_header(fname):
    with fits.open(fname) as hdu:
        primhdr = hdu[0].header
        if len(hdu) > 1:
            imghdr = hdu[1].header
            primhdr.update(imghdr)
    if primhdr['INSTRUME'] != 'EFOSC':
        print("[WARNING] - FITS file not originating from NTT/EFOSC!")
    return primhdr


def create_pixel_array(hdr, dispaxis):
    """Load reference array from header using CRVAL, CDELT, CRPIX along dispersion axis"""
    if dispaxis not in [1, 2]:
        raise ValueError("Dispersion Axis must be 1 (X-axis) or 2 (Y-axis)!")
    p = hdr['CRVAL%i' % dispaxis]
    s = hdr['CDELT%i' % dispaxis]
    r = hdr['CRPIX%i' % dispaxis]
    N = hdr['NAXIS%i' % dispaxis]
    # -- If data are from NOT then check for binning and rescale CRPIX:
    pix_array = p + s*(np.arange(N) - (r - 1))
    return pix_array


def get_binning(fname):
    hdr = fits.getheader(fname)
    ccd_setup = get_binning_from_hdr(hdr)
    return ccd_setup


def get_binning_from_hdr(hdr):
    binx = hdr['ESO DET WIN1 BINX']
    biny = hdr['ESO DET WIN1 BINY']
    read = hdr['ESO DET READ SPEED'].strip()
    ccd_setup = "%ix%i_%i" % (binx, biny, read)
    return ccd_setup


def get_filter(hdr):
    filter_name = hdr['ESO INS FILT1 NAME']
    if filter_name == 'Free':
        return "Open"
    return filter_name

def get_grism(hdr):
    grism_name = grism_translate(hdr['ESO INS GRIS1 NAME'])
    return grism_name

def get_slit(hdr):
    # HIERARCH ESO INS SLIT1 NAME    = 'slit#1.2' / Slit common name.
    slit_name = slit_translate(hdr['ESO INS SLIT1 NAME'])
    return slit_name

def get_airmass(hdr):
    """Return the average airmass at mid-exposure"""
    airm_start = hdr['ESO TEL AIRM START']
    airm_end = hdr['ESO TEL AIRM END']
    airmass = 0.5*(airm_start + airm_end)
    return airmass



#Example header:
# SIMPLE  =                    T / Written by IDL:  Sat Mar 17 04:23:05 2012
# BITPIX  =                  -32 / Bits per pixel
# NAXIS   =                    1 / Number of axes
# NAXIS1  =                 1030 /Number of positions along axis 1
# EXTEND  =                    F / File may contain extensions
# IRAF-TLM= '2012-02-02T07:27:58' / Time of last modification
# OBJECT  = 'CQ0831+0930'        /
# DATE    = '2012-02-02T07:27:58'
# IRAF-MAX=           0.000000E0  /  DATA MAX
# IRAF-MIN=           0.000000E0  /  DATA MIN
# ORIGIN  = 'NOAO-IRAF FITS Image Kernel July 2003' / FITS file originator
# DATE    = '2012-02-02T07:23:39' / Date FITS file was generated
# COMMENT NOST 100-2.0: Hanisch,R. et al. 2001, Astron. & Astrophys. 376, 559
#
#
# CRPIX1  =                   1. / Reference pixel
# CRVAL1  =      3804.7008200246 / Coordinate at reference pixel
# CTYPE1  = 'LINEAR  '           / Units of coordinate
# CTYPE2  = 'LINEAR  '           / Units of coordinate
# BUNIT   = 'erg/cm2/s/A'        / Units of data values
#
# MIDASFTP= 'IMAGE   '           / MIDAS File Type
#
# RA      =           127.927598 / MIDAS desc.: O_POS(1)
# DEC     =               9.5059 / MIDAS desc.: O_POS(2)
# EQUINOX =                2000. / MIDAS desc.: O_POS(3)
# DATE-OBS= '2011-11-22'         / MIDAS desc.: O_TIME(1)
# MJD-OBS =         55887.347975 / MIDAS desc.: O_TIME(4)
# TM-START=       30065.04000099 / MIDAS desc.: O_TIME(5)
# EXPTIME =             900.0005 / MIDAS desc.: O_TIME(7)
# TELESCOP= 'ESO-NTT '           / MIDAS desc.: TELESCOP(1)
# INSTRUME= 'EFOSC   '           / MIDAS desc.: INSTRUME(1)
# RADECSYS= 'FK5     '           / MIDAS desc.: RADECSYS(1)
# OBSERVER= 'UNKNOWN '           / MIDAS desc.: OBSERVER(1)
# HISTORY Converted from: crrdatar0019.fits                                      \
# HISTORY          REBIN/ROTATE spec1 rot R1 + NO                                \
# HISTORY
# UTC     =               30057. / 08:20:57.000 UTC at start (sec)
# LST     =              27715.8 / 07:41:55.800 LST at start (sec)
# PI-COI  = 'UNKNOWN '           / PI-COI name.
# CD1_1   =     4.11520195625661 / Translation matrix element.
# CD2_2   =                   1. / Translation matrix element.
# ORIGFILE= 'EFOSC_Spectrum326_0020.fits' / Original File Name
# ARCFILE = 'EFOSC.2011-11-22T08:21:05.040.fits' / Archive File Name
# CHECKSUM= 'fe1Tie1Rfe1Rfe1R'   / HDU checksum updated 2012-03-17T03:23:05
# DATASUM = '1362823054'         / data unit checksum updated 2012-03-17T03:23:05
# O_BZERO =               32768. / Original BZERO Value
# HIERARCH ESO ADA ABSROT END    =  -136.6974 / Abs rot angle at exp end (deg)
# HIERARCH ESO ADA ABSROT START  = -140.79743 / Abs rot angle at exp start (deg)
# HIERARCH ESO ADA GUID STATUS   = 'OFF     ' / Status of autoguider
# HIERARCH ESO ADA POSANG=    -71.698 / Position angle at start
# HIERARCH ESO DET BITS  =         16 / Bits per pixel readout
# HIERARCH ESO DET CHIP1 DATE    = '2000-08-30' / Date of installation [YYYY-MM-DD
# HIERARCH ESO DET CHIP1 ID      = 'ccd40   ' / Detector chip identification
# HIERARCH ESO DET CHIP1 INDEX   =          1 / Chip index
# HIERARCH ESO DET CHIP1 NAME    = 'LORAL   ' / Detector chip name
# HIERARCH ESO DET CHIP1 NX      =       2048 / # of pixels along X
# HIERARCH ESO DET CHIP1 NY      =       2048 / # of pixels along Y
# HIERARCH ESO DET CHIP1 PSZX    =        15. / Size of pixel in X
# HIERARCH ESO DET CHIP1 PSZY    =        15. / Size of pixel in Y
# HIERARCH ESO DET CHIP1 X       =          1 / X location in array
# HIERARCH ESO DET CHIP1 XGAP    =         0. / Gap between chips along x
# HIERARCH ESO DET CHIP1 Y       =          1 / Y location in array
# HIERARCH ESO DET CHIP1 YGAP    =         0. / Gap between chips along y
# HIERARCH ESO DET CHIPS =          1 / # of chips in detector array
# HIERARCH ESO DET DATE  = '2000-08-30' / Installation date
# HIERARCH ESO DET DEC   =         0. / Apparent 00:00:00.0 DEC at start
# HIERARCH ESO DET DID   = 'ESO-VLT-DIC.CCDDCS,ESO-VLT-DIC.FCDDCS' / Diction
# HIERARCH ESO DET EXP NO=       5401 / Unique exposure ID number
# HIERARCH ESO DET EXP RDTTIME   =      9.711 / image readout time
# HIERARCH ESO DET EXP TYPE      = 'Normal  ' / Exposure type
# HIERARCH ESO DET EXP XFERTIM   =     22.315 / image transfer time
# HIERARCH ESO DET FRAM ID       =          1 / Image sequencial number
# HIERARCH ESO DET FRAM TYPE     = 'Normal  ' / Type of frame
# HIERARCH ESO DET ID    = 'CCD FIERA - Rev: 3.96' / Detector system Id
# HIERARCH ESO DET NAME  = 'efosc - ccdefosc' / Name of detector system
# HIERARCH ESO DET OUT1 CHIP     =          1 / Chip to which the output belongs
# HIERARCH ESO DET OUT1 CONAD    =        1.1 / Conversion from ADUs to electrons
# HIERARCH ESO DET OUT1 GAIN     =       0.91 / Conversion from electrons to ADU
# HIERARCH ESO DET OUT1 ID       = 'L       ' / Output ID as from manufacturer
# HIERARCH ESO DET OUT1 INDEX    =          1 / Output index
# HIERARCH ESO DET OUT1 NAME     = 'L       ' / Description of output
# HIERARCH ESO DET OUT1 NX       =       1024 / valid pixels along X
# HIERARCH ESO DET OUT1 NY       =       1024 / valid pixels along Y
# HIERARCH ESO DET OUT1 OVSCX    =          0 / Overscan region in X
# HIERARCH ESO DET OUT1 OVSCY    =          6 / Overscan region in Y
# HIERARCH ESO DET OUT1 PRSCX    =          6 / Prescan region in X
# HIERARCH ESO DET OUT1 PRSCY    =          0 / Prescan region in Y
# HIERARCH ESO DET OUT1 RON      =        7.8 / Readout noise per output (e-)
# HIERARCH ESO DET OUT1 X=       2048 / X location of output
# HIERARCH ESO DET OUT1 Y=          1 / Y location of output
# HIERARCH ESO DET OUTPUTS       =          1 / # of outputs
# HIERARCH ESO DET OUTREF=          0 / reference output
# HIERARCH ESO DET RA    =         0. / Apparent 00:00:00.0 RA at start
# HIERARCH ESO DET READ CLOCK    = 'read L port Fast' / Readout clock pattern used
# HIERARCH ESO DET READ MODE     = 'normal  ' / Readout method
# HIERARCH ESO DET READ NFRAM    =          1 / Number of readouts buffered in sin
# HIERARCH ESO DET READ SPEED    = 'fastL   ' / Readout speed
# HIERARCH ESO DET SHUT ID       = 'ccd shutter' / Shutter unique identifier
# HIERARCH ESO DET SHUT TMCLOS   =      0.065 / Time taken to close shutter
# HIERARCH ESO DET SHUT TMOPEN   =      0.064 / Time taken to open shutter
# HIERARCH ESO DET SHUT TYPE     = 'Slit    ' / type of shutter
# HIERARCH ESO DET SOFW MODE     = 'Normal  ' / CCD sw operational mode
# HIERARCH ESO DET WIN1 BINX     =          2 / Binning factor along X
# HIERARCH ESO DET WIN1 BINY     =          2 / Binning factor along Y
# HIERARCH ESO DET WIN1 DIT1     = 900.000549 / actual subintegration time
# HIERARCH ESO DET WIN1 DKTM     =   900.1025 / Dark current time
# HIERARCH ESO DET WIN1 NDIT     =          1 / # of subintegrations
# HIERARCH ESO DET WIN1 NX       =       1030 / # of pixels along X
# HIERARCH ESO DET WIN1 NY       =       1030 / # of pixels along Y
# HIERARCH ESO DET WIN1 ST       = T / If T, window enabled
# HIERARCH ESO DET WIN1 STRX     =          1 / Lower left pixel in X
# HIERARCH ESO DET WIN1 STRY     =          1 / Lower left pixel in Y
# HIERARCH ESO DET WIN1 UIT1     =       900. / user defined subintegration time
# HIERARCH ESO DET WINDOWS       =          1 / # of windows readout
# HIERARCH ESO DPR CATG  = 'SCIENCE ' / Observation category
# HIERARCH ESO DPR TECH  = 'SPECTRUM' / Observation technique
# HIERARCH ESO DPR TYPE  = 'OBJECT  ' / Observation type
# HIERARCH ESO INS DATE  = '2000-06-16' / Instrument release date (yyyy-mm-d
# HIERARCH ESO INS DID   = 'ESO-VLT-DIC.EFOSC_ICS-1.21' / Data dictionary fo
# HIERARCH ESO INS DPOR POS      =         0. / Position move
# HIERARCH ESO INS DPOR ST       = F / Instrument depolarizer rotating.
# HIERARCH ESO INS FILT1 ID      = 'F1      ' / Filter unique id.
# HIERARCH ESO INS FILT1 NAME    = 'Free    ' / Filter name.
# HIERARCH ESO INS FILT1 NO      =          1 / Filter wheel position index.
# HIERARCH ESO INS GRIS1 ID      = 'F12     ' / OPTIi unique ID.
# HIERARCH ESO INS GRIS1 NAME    = 'Gr#6    ' / OPTIi name.
# HIERARCH ESO INS GRIS1 NO      =         12 / OPTIi slot number.
# HIERARCH ESO INS GRIS1 TYPE    = 'FILTER  ' / OPTIi element.
# HIERARCH ESO INS ID    = 'EFOSC/1.57' / Instrument ID.
# HIERARCH ESO INS MODE  = 'DEFAULT ' / Instrument mode used.
# HIERARCH ESO INS MOS1 LEN      =      1777. / MOSi slit length [arcsec]
# HIERARCH ESO INS MOS1 POSX     =      1002. / MOSi slit position [pix]
# HIERARCH ESO INS MOS1 POSY     =       992. / MOSi slit position [pix]
# HIERARCH ESO INS SLIT1 ENC     =     257870 / Slit absolute position [Enc].
# HIERARCH ESO INS SLIT1 LEN     =         0. / SLIT length [arcsec].
# HIERARCH ESO INS SLIT1 NAME    = 'slit#1.2' / Slit common name.
# HIERARCH ESO INS SLIT1 NO      =          5 / Slide position.
# HIERARCH ESO INS SLIT1 WID     =         0. / SLIT width [arcsec].
# HIERARCH ESO INS SWSIM = 'NORMAL  ' / Software simulation.
# HIERARCH ESO INS WP NAME       = 'HALF    ' / WP name, valid values are HALF/QUA
# HIERARCH ESO INS WP ST = 'OUT     ' / WP position.
# HIERARCH ESO OBS DID   = 'ESO-VLT-DIC.OBS-1.11' / OBS Dictionary
# HIERARCH ESO OBS EXECTIME      =          0 / Expected execution time
# HIERARCH ESO OBS GRP   = '0       ' / linked blocks
# HIERARCH ESO OBS ID    =  100298255 / Observation block ID
# HIERARCH ESO OBS NAME  = 'CQ0831+0930' / OB name
# HIERARCH ESO OBS OBSERVER      = 'UNKNOWN ' / Observer Name
# HIERARCH ESO OBS PI-COI ID     =      75259 / ESO internal PI-COI ID
# HIERARCH ESO OBS PI-COI NAME   = 'UNKNOWN ' / PI-COI name
# HIERARCH ESO OBS PROG ID       = '088.A-0098(A)' / ESO program identification
# HIERARCH ESO OBS START = '2011-11-22T08:16:26' / OB start time
# HIERARCH ESO OBS TARG NAME     = 'CQ0831+0930' / OB target name
# HIERARCH ESO OBS TPLNO =          2 / Template number within OB
# HIERARCH ESO OCS CON WCSFITS   = T / Setup OS to add WCS in fits  .
# HIERARCH ESO OCS DET1 IMGNAME  = 'EFOSC_Spectrum' / Data File Name.
# HIERARCH ESO TEL AIRM END      =      1.297 / Airmass at end
# HIERARCH ESO TEL AIRM START    =      1.316 / Airmass at start
# HIERARCH ESO TEL ALT   =     49.423 / Alt angle at start (deg)
# HIERARCH ESO TEL AMBI FWHM END =       0.57 / Observatory Seeing queried from AS
# HIERARCH ESO TEL AMBI FWHM START       =       0.67 / Observatory Seeing queried
# HIERARCH ESO TEL AMBI PRES END =      768.6 / Observatory ambient air pressure q
# HIERARCH ESO TEL AMBI PRES START       =      768.6 / Observatory ambient air pr
# HIERARCH ESO TEL AMBI RHUM     =        15. / Observatory ambient relative humi
# HIERARCH ESO TEL AMBI TEMP     =       15.6 / Observatory ambient temperature qu
# HIERARCH ESO TEL AMBI WINDDIR  =        35. / Observatory ambient wind directio
# HIERARCH ESO TEL AMBI WINDSP   =        4.8 / Observatory ambient wind speed que
# HIERARCH ESO TEL AZ    =    199.333 / Az angle at start (deg) S=0,W=90
# HIERARCH ESO TEL CHOP ST       = F / True when chopping is active
# HIERARCH ESO TEL DATE  = '2009-06-05T01:00:00' / TCS installation date
# HIERARCH ESO TEL DID   = 'ESO-VLT-DIC.TCS' / Data dictionary for TEL
# HIERARCH ESO TEL DOME STATUS   = 'FULLY-OPEN' / Dome status
# HIERARCH ESO TEL FOCU ID       = 'NB      ' / Telescope focus station ID
# HIERARCH ESO TEL FOCU LEN      =     38.501 / Focal length (m)
# HIERARCH ESO TEL FOCU SCALE    =       5.36 / Focal scale (arcsec/mm)
# HIERARCH ESO TEL FOCU VALUE    =      -3.38 / M2 setting (mm)
# HIERARCH ESO TEL GEOELEV       =      2377. / Elevation above sea level (m)
# HIERARCH ESO TEL GEOLAT=   -29.2584 / Tel geo latitute (+=North) (deg)
# HIERARCH ESO TEL GEOLON=   -70.7345 / Tel geo longitude (+=East) (deg)
# HIERARCH ESO TEL ID    = 'v 8.27+ ' / TCS version number
# HIERARCH ESO TEL MOON DEC      =  -11.00212 / -11:00:07.6 DEC (J2000) (deg)
# HIERARCH ESO TEL MOON RA       = 196.753158 / 13:07:00.7 RA (J2000) (deg)
# HIERARCH ESO TEL OPER  = 'I. CONDOR' / Telescope Operator
# HIERARCH ESO TEL PARANG END    =   -168.034 / Parallactic angle at end (deg)
# HIERARCH ESO TEL PARANG START  =   -162.976 / Parallactic angle at start (deg)
# HIERARCH ESO TEL TARG ALPHA    =   83142.36 / Alpha coordinate for the target
# HIERARCH ESO TEL TARG COORDTYPE= 'M       ' / Coordinate type (M=mean A=apparent
# HIERARCH ESO TEL TARG DELTA    =    93029.8 / Delta coordinate for the target
# HIERARCH ESO TEL TARG EPOCH    =      2000. / Epoch
# HIERARCH ESO TEL TARG EPOCHSYSTEM      = 'J       ' / Epoch system (default J=Ju
# HIERARCH ESO TEL TARG EQUINOX  =      2000. / Equinox
# HIERARCH ESO TEL TARG PARALLAX =         0. / Parallax
# HIERARCH ESO TEL TARG PMA      =         0. / Proper Motion Alpha
# HIERARCH ESO TEL TARG PMD      =         0. / Proper motion Delta
# HIERARCH ESO TEL TARG RADVEL   =         0. / Radial velocity
# HIERARCH ESO TEL TH M1 TEMP    =        -1. / M1 superficial temperature
# HIERARCH ESO TEL TRAK STATUS   = 'NORMAL  ' / Tracking status
# HIERARCH ESO TEL TSS TEMP8     =      14.38 / Temperature Sensing System
# HIERARCH ESO TPL DID   = 'ESO-VLT-DIC.TPL-1.9' / Data dictionary for TPL
# HIERARCH ESO TPL EXPNO =          1 / Exposure number within template
# HIERARCH ESO TPL ID    = 'EFOSC_spec_obs_Spectrum' / Template signature ID
# HIERARCH ESO TPL NAME  = 'Spectroscopy' / Template name
# HIERARCH ESO TPL NEXP  =          1 / Number of exposures within templat
# HIERARCH ESO TPL PRESEQ= 'EFOSC_spec_obs_Spectrum.seq' / Sequencer script
# HIERARCH ESO TPL START = '2011-11-22T08:20:40' / TPL start time
# HIERARCH ESO TPL VERSION       = '1.0     ' / Version of the template
#
# HISTORY  ESO-DESCRIPTORS START   ................
# HISTORY  'LHCUTS','R*4',1,4,'5E14.7'
# HISTORY   0.0000000E+00 0.0000000E+00 1.1754944E-38 7.8100000E+02
# HISTORY
# HISTORY  'ROTANG_FROM_CD-MATRIX','R*8',1,2,'3E23.15'
# HISTORY    1.251366203504492E+00  1.251366203504492E+00
# HISTORY
# HISTORY  ESO-DESCRIPTORS END     ................
# BANDID1 = 'spectrum - background average, weights variance, clean no'
# BANDID2 = 'raw - background average, weights none, clean no'
# BANDID3 = 'background - background average'
# BANDID4 = 'sigma - background average, weights variance, clean no'
# APNUM1  = '1 1 481.35 490.46'
# WCSDIM  =                    3
# CTYPE3  = 'LINEAR  '
# CD3_3   =                   1.
# LTM1_1  =                   1.
# LTM2_2  =                   1.
# LTM3_3  =                   1.
# WAT0_001= 'system=equispec'
# WAT1_001= 'wtype=linear label=Wavelength units=angstroms'
# WAT2_001= 'wtype=linear'
# WAT3_001= 'wtype=linear'
# DC-FLAG =                    0
# DCLOG1  = 'REFSPEC1 = arc_1d.fits'
# AIRMASS =                1.297
# EX-FLAG =                    0
# CA-FLAG =                    0



"""
Automatically Classify and Reduce a given Data Set
"""

from astropy.io import fits
from collections import defaultdict
import os
import sys
import datetime

from pynot import alfosc
from pynot.data import io
from pynot.data import organizer as do
from pynot.calibs import combine_bias_frames, combine_flat_frames, normalize_spectral_flat
from pynot.extraction import auto_extract
from pynot import extract_gui
from pynot.functions import get_options, get_version_number
from pynot.wavecal import rectify, WavelengthError
from pynot.identify_gui import create_pixtable
from pynot.scired import raw_correction, auto_fit_background, correct_cosmics, correct_raw_file
from pynot.response import calculate_response, flux_calibrate

from PyQt5.QtWidgets import QApplication

code_dir = os.path.dirname(os.path.abspath(__file__))
calib_dir = os.path.join(code_dir, 'calib/')
defaults_fname = os.path.join(calib_dir, 'default_options.yml')
__version__ = get_version_number()


class Report(object):
    def __init__(self, verbose=False):
        self.verbose = verbose
        self.time = datetime.datetime.now()
        self.fname = 'pynot_%s.log' % self.time.strftime('%d%b%Y-%Hh%Mm%S')
        self.remarks = list()
        self.lines = list()
        self.header = """
        #  PyNOT Data Processing Pipeline
        # ================================
        # version %s
        %s

        """ % (__version__, self.time.strftime("%b %d, %Y  %H:%M:%S"))
        self.report = ""

        if self.verbose:
            print(self.header)

    def clear(self):
        self.lines = list()
        self.remarks = list()

    def set_filename(self, fname):
        self.fname = fname

    def commit(self, text):
        if self.verbose:
            print(text, end='', flush=True)
        self.lines.append(text)

    def error(self, text):
        text = ' [ERROR]  - ' + text
        if self.verbose:
            print(text)
        if text[-1] != '\n':
            text += '\n'
        self.lines.append(text)

    def warn(self, text):
        text = '[WARNING] - ' + text
        if self.verbose:
            print(text)
        if text[-1] != '\n':
            text += '\n'
        self.lines.append(text)

    def write(self, text, prefix='          - '):
        text = prefix + text
        if self.verbose:
            print(text)
        if text[-1] != '\n':
            text += '\n'
        self.lines.append(text)

    def add_linebreak(self):
        if self.verbose:
            print("")
        self.lines.append("\n")

    def add_remark(self, text):
        self.remarks.append(text)

    def _make_report(self):
        remark_str = ''.join(self.remarks)
        lines_str = ''.join(self.lines)
        self.report = '\n'.join([self.header, remark_str, lines_str])

    def print_report(self):
        self._make_report()
        print(self.report)

    def save(self):
        self._make_report()
        with open(self.fname, 'w') as output:
            output.write(self.report)

    def exit(self):
        print(" - Pipeline terminated.")
        print(" Consult the log: %s\n" % self.fname)
        self.save()

    def fatal_error(self):
        print(" !! FATAL ERROR !!")
        print(" Consult the log: %s\n" % self.fname)
        self.save()


class State(dict):
    """A collection of variables for the pipeline, such as arc line ID tables etc."""
    def __init__(self):
        dict.__init__(self, {})
        self.current = ''

    def print_current_state(self):
        print(self.current)

    def set_current_state(self, state):
        self.current = state



def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False):
    log = Report(verbose)
    status = State()

    global app
    app = QApplication(sys.argv)

    # -- Parse Options from YAML
    options = get_options(defaults_fname)

    user_options = get_options(options_fname)
    for section_name, section in user_options.items():
        if isinstance(section, dict):
            options[section_name].update(section)
        else:
            options[section_name] = section

    if object_id is None:
        pass
    elif isinstance(object_id, str):
        object_id = [object_id]
    elif isinstance(object_id, list):
        if isinstance(object_id[0], str):
            pass
        else:
            log.error("Wrong input type for `object id`. Must be list of strings")
            log.error("not list of %r" % type(object_id[0]))
            log.fatal_error()
            return
    else:
        log.error("Wrong input type for `object id`. Must be string or list of strings")
        log.error("not %r" % type(object_id))
        log.fatal_error()
        return

    if interactive:
        # Set all interactive steps to True
        options['identify']['interactive'] = True
        options['identify']['all'] = True
        options['extract']['interactive'] = True
        options['response']['interactive'] = True

    dataset_fname = options['dataset']
    if dataset_fname and os.path.exists(dataset_fname):
        # -- load collection
        database = io.load_database(dataset_fname)
        log.write("Loaded file classification database: %s" % dataset_fname)
        # -- reclassify (takes already identified files into account)

    else:
        log.error("Dataset does not exist : %s" % dataset_fname)
        log.fatal_error()
        return

    # -- Organize object files in dataset:
    if 'SPEC_OBJECT' not in database:
        log.error("No spectroscopic data found in the dataset!")
        log.error("Check the classification table... object type 'SPEC_OBJECT' missing")
        log.fatal_error()
        return
    object_filelist = database['SPEC_OBJECT']
    object_images = list(map(do.RawImage, object_filelist))

    log.add_linebreak()
    log.write(" - The following objects were found in the dataset:", prefix='')
    log.write("      OBJECT           GRISM        SLIT      EXPTIME       FILENAME", prefix='')
    for sci_img in object_images:
        output_variables = (sci_img.object, sci_img.grism, sci_img.slit, sci_img.exptime, os.path.basename(sci_img.filename))
        log.write("%20s  %9s  %11s   %5.0f  %s" % output_variables, prefix='')
    log.add_linebreak()

    # get list of unique grisms in dataset:
    grism_list = list()
    for sci_img in object_images:
        grism_name = alfosc.grism_translate[sci_img.grism]
        if grism_name not in grism_list:
            grism_list.append(grism_name)

    # -- Check arc line files:
    arc_images = list()
    # for arc_type in ['ARC_He', 'ARC_HeNe', 'ARC_Ne', 'ARC_ThAr']:
    for arc_type in ['ARC_HeNe', 'ARC_ThAr']:
        # For now only HeNe arc lines are accepted!
        # Implement ThAr and automatic combination of He + Ne
        if arc_type in database.keys():
            arc_images += database[arc_type]

    if len(arc_images) == 0:
        log.error("No arc line calibration data found in the dataset!")
        log.error("Check the classification table... object type 'ARC_HeNe' or 'ARC_ThAr' missing")
        log.fatal_error()
        return

    arc_images_for_grism = defaultdict(list)
    for arc_img in arc_images:
        raw_grism = fits.getheader(arc_img)['ALGRNM']
        this_grism = alfosc.grism_translate[raw_grism]
        arc_images_for_grism[this_grism].append(arc_img)

    for grism_name in grism_list:
        if len(arc_images_for_grism[grism_name]) == 0:
            log.error("No arc frames defined for grism: %s" % grism_name)
            log.fatal_error()
            return
        else:
            log.write("%s has necessary arc files." % grism_name)

    log.add_linebreak()
    identify_all = options['identify']['all']
    identify_interactive = options['identify']['interactive']
    if identify_interactive and identify_all:
        grisms_to_identify = []
        log.write("Identify: interactively reidentify arc lines for all objects")
        log.add_linebreak()

    elif identify_interactive and not identify_all:
        # Make pixeltable for all grisms:
        grisms_to_identify = grism_list
        log.write("Identify: interactively identify all grisms in dataset:")
        log.write(", ".join(grisms_to_identify))
        log.add_linebreak()

    else:
        # Check if pixeltables exist:
        grisms_to_identify = []
        for grism_name in grism_list:
            pixtab_fname = os.path.join(calib_dir, '%s_pixeltable.dat' % grism_name)
            if not os.path.exists(pixtab_fname):
                log.write("%s : pixel table does not exist. Will identify lines..." % grism_name)
                grisms_to_identify.append(grism_name)
            else:
                log.write("%s : pixel table already exists" % grism_name)
                status[grism_name+'_pixtab'] = pixtab_fname
                status[pixtab_fname] = options['identify']['order_wl']
        log.add_linebreak()


    # Identify interactively for grisms that are not defined
    # add the new pixel tables to the calib cache for future use
    for grism_name in grisms_to_identify:
        log.write("Starting interactive definition of pixel table for %s" % grism_name)
        try:
            arc_fname = arc_images_for_grism[grism_name][0]
            if grism_name+'_pixtab' in options:
                pixtab_fname = options[grism_name+'_pixtab']
            else:
                pixtab_fname = os.path.join(calib_dir, '%s_pixeltable.dat' % grism_name)
            linelist_fname = ''
            log.write("Input arc line frame: %s" % arc_fname)
            poly_order, saved_pixtab_fname, msg = create_pixtable(arc_fname, grism_name,
                                                                  pixtab_fname, linelist_fname,
                                                                  order_wl=options['identify']['order_wl'],
                                                                  app=app)
            status[saved_pixtab_fname] = poly_order
            status[grism_name+'_pixtab'] = saved_pixtab_fname
            log.commit(msg)
        except:
            log.error("Identification of arc lines failed!")
            log.fatal_error()
            log.save()
            print("Unexpected error:", sys.exc_info()[0])
            raise


    # Save overview log:
    print("")
    print(" - Pipeline setup ended successfully.")
    print("   Consult the overview log: %s\n\n" % log.fname)
    log.save()


    # ------------------------------------------------------------------
    # -- Start Main Reduction:
    if object_id is None:
        # Loop over all:
        objects_to_reduce = object_images
    else:
        objects_to_reduce = list()
        for img in object_images:
            if img.object in object_id:
                objects_to_reduce.append(img)

        if len(objects_to_reduce) == 0:
            log.error("No data matched the given object ID: %r" % object_id)
            log.fatal_error()
            return

    for sci_img in objects_to_reduce:
        # Create working directory:
        raw_base = os.path.basename(sci_img.filename).split('.')[0][2:]
        output_dir = sci_img.target_name + '_' + raw_base
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        # Start new log in working directory:
        log_fname = os.path.join(output_dir, 'pynot.log')
        log.clear()
        log.set_filename(log_fname)
        log.write("Starting PyNOT Longslit Spectroscopic Reduction")
        log.add_linebreak()
        log.write("Target Name: %s" % sci_img.target_name)
        log.write("Input Filename: %s" % sci_img.filename)
        log.write("Saving output to directory: %s" % output_dir)

        # Prepare output filenames:
        grism = alfosc.grism_translate[sci_img.grism]
        master_bias_fname = os.path.join(output_dir, 'MASTER_BIAS.fits')
        comb_flat_fname = os.path.join(output_dir, 'FLAT_COMBINED_%s_%s.fits' % (grism, sci_img.slit))
        norm_flat_fname = os.path.join(output_dir, 'NORM_FLAT_%s_%s.fits' % (grism, sci_img.slit))
        rect2d_fname = os.path.join(output_dir, 'RECT2D_%s.fits' % (sci_img.target_name))
        bgsub2d_fname = os.path.join(output_dir, 'BGSUB2D_%s.fits' % (sci_img.target_name))
        response_pdf = os.path.join(output_dir, 'RESPONSE_%s.pdf' % (grism))
        corrected_2d_fname = os.path.join(output_dir, 'CORRECTED2D_%s.fits' % (sci_img.target_name))
        flux2d_fname = os.path.join(output_dir, 'FLUX2D_%s.fits' % (sci_img.target_name))
        flux1d_fname = os.path.join(output_dir, 'FLUX1D_%s.fits' % (sci_img.target_name))
        extract_pdf_fname = os.path.join(output_dir, 'extract1D_details.pdf')

        # Combine Bias Frames matched for CCD setup:
        bias_frames = sci_img.match_files(database['BIAS'], date=False)
        if options['mbias']:
            master_bias_fname = options['mbias']
            log.write("Using static master bias frame: %s" % options['mbias'])
        elif len(bias_frames) < 3:
            log.error("Must have at least 3 bias frames to combine, not %i" % len(bias_frames))
            log.error("otherwise provide a static 'master bias' frame!")
            log.fatal_error()
            return
        else:
            log.write("Running task: Bias Combination")
            try:
                _, bias_msg = combine_bias_frames(bias_frames, output=master_bias_fname,
                                                  kappa=options['bias']['kappa'],
                                                  method=options['bias']['method'],
                                                  overwrite=True)
                log.commit(bias_msg)
                log.add_linebreak()
                status['master_bias'] = master_bias_fname
            except:
                log.error("Median combination of bias frames failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise


        # Combine Flat Frames matched for CCD setup, grism, slit and filter:
        flat_frames = sci_img.match_files(database['SPEC_FLAT'], date=False, grism=True, slit=True, filter=True)
        if options['mflat']:
            if options['mflat'] is None:
                norm_flat_fname = ''
            elif options['mflat'].lower() in ['none', 'null']:
                norm_flat_fname = ''
            else:
                norm_flat_fname = options['mflat']
            log.write("Using static master flat frame: %s" % options['mflat'])
        elif len(flat_frames) == 0:
            log.error("No flat frames provided!")
            log.fatal_error()
            return
        else:
            try:
                log.write("Running task: Spectral Flat Combination")
                _, flat_msg = combine_flat_frames(flat_frames, comb_flat_fname, mbias=master_bias_fname,
                                                  kappa=options['flat']['kappa'],
                                                  method=options['flat']['method'], overwrite=True,
                                                  mode='spec', dispaxis=sci_img.dispaxis)
                log.commit(flat_msg)
                log.add_linebreak()
                status['flat_combined'] = comb_flat_fname
            except ValueError as err:
                log.commit(str(err)+'\n')
                log.fatal_error()
                return
            except:
                log.error("Combination of flat frames failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise

            # Normalize the spectral flat field:
            try:
                log.write("Running task: Spectral Flat Normalization")
                _, norm_msg = normalize_spectral_flat(comb_flat_fname, output=norm_flat_fname,
                                                      fig_dir=output_dir, dispaxis=sci_img.dispaxis,
                                                      **options['flat'])
                log.commit(norm_msg)
                log.add_linebreak()
                status['master_flat'] = norm_flat_fname
            except:
                log.error("Normalization of flat frames failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise


        # Identify lines in arc frame:
        arc_fname, = sci_img.match_files(arc_images, date=False, grism=True, slit=True, filter=True, get_closest_time=True)
        corrected_arc2d_fname = os.path.join(output_dir, 'corr_arc2d.fits')
        log.write("Running task: Bias and Flat Field Correction of Arc Frame")
        try:
            output_msg = correct_raw_file(arc_fname, bias_fname=master_bias_fname,
                                          output=corrected_arc2d_fname, overwrite=True, overscan=50)
            log.commit(output_msg)
            log.add_linebreak()
        except:
            log.error("Bias and flat field correction of Arc frame failed!")
            log.fatal_error()
            print("Unexpected error:", sys.exc_info()[0])
            raise


        if grism+'_pixtab' in options:
            pixtab_fname = options[grism_name+'_pixtab']
        else:
            pixtab_fname = os.path.join(calib_dir, '%s_pixeltable.dat' % grism)

        if identify_interactive and identify_all:
            log.write("Running task: Arc Line Identification")
            try:
                linelist_fname = ''
                order_wl, pixtable, msg = create_pixtable(corrected_arc2d_fname, grism,
                                                          pixtab_fname, linelist_fname,
                                                          order_wl=options['identify']['order_wl'],
                                                          app=app)
                status[pixtable] = order_wl
                status[grism+'_pixtab'] = pixtable
                log.commit(msg)
                log.add_linebreak()
            except:
                log.error("Identification of arc lines failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise
        else:
            # -- or use previous line identifications
            pixtable = status[grism+'_pixtab']
            order_wl = status[pixtable]


        # Response Function:
        if 'SPEC_FLUX-STD' in database:
            flux_std_files = sci_img.match_files(database['SPEC_FLUX-STD'],
                                                 date=False, grism=True, slit=True, filter=True, get_closest_time=True)
        else:
            flux_std_files = []

        if len(flux_std_files) == 0:
            log.warn("No spectroscopic standard star was found in the dataset!")
            log.warn("The reduced spectra will not be flux calibrated")
            status['RESPONSE'] = None

        else:
            std_fname = flux_std_files[0]
            # response_fname = os.path.join(output_dir, 'response_%s.fits' % (grism))
            response_fname = 'response_%s.fits' % (grism)
            if os.path.exists(response_fname) and not options['response']['force']:
                log.write("Response function already exists: %s" % response_fname)
                log.add_linebreak()
                status['RESPONSE'] = response_fname
            else:
                std_fname = flux_std_files[0]
                log.write("Running task: Calculation of Response Function")
                log.write("Spectroscopic Flux Standard: %s" % std_fname)
                try:
                    response_fname, response_msg = calculate_response(std_fname, arc_fname=corrected_arc2d_fname,
                                                                      pixtable_fname=pixtable,
                                                                      bias_fname=master_bias_fname,
                                                                      flat_fname=norm_flat_fname,
                                                                      output=response_fname,
                                                                      output_dir=output_dir, pdf_fname=response_pdf,
                                                                      order=options['response']['order'],
                                                                      interactive=options['response']['interactive'],
                                                                      dispaxis=sci_img.dispaxis, order_wl=order_wl,
                                                                      order_bg=options['skysub']['order_bg'],
                                                                      rectify_options=options['rectify'],
                                                                      app=app)
                    status['RESPONSE'] = response_fname
                    log.commit(response_msg)
                    log.add_linebreak()
                except:
                    log.error("Calculation of response function failed!")
                    print("Unexpected error:", sys.exc_info()[0])
                    status['RESPONSE'] = ''
                    log.warn("No flux calibration will be performed!")
                    log.add_linebreak()


        # Bias correction, Flat correction
        log.write("Running task: Bias and Flat Field Correction")
        try:
            output_msg = raw_correction(sci_img.data, sci_img.header, master_bias_fname, norm_flat_fname,
                                        output=corrected_2d_fname, overwrite=True, overscan=50)
            log.commit(output_msg)
            log.add_linebreak()
        except:
            log.error("Bias and flat field correction failed!")
            log.fatal_error()
            print("Unexpected error:", sys.exc_info()[0])
            raise


        # Call rectify
        log.write("Running task: 2D Rectification and Wavelength Calibration")
        try:
            rect_msg = rectify(corrected_2d_fname, corrected_arc2d_fname, pixtable, output=rect2d_fname, fig_dir=output_dir,
                               dispaxis=sci_img.dispaxis, order_wl=order_wl, **options['rectify'])
            log.commit(rect_msg)
            log.add_linebreak()
        except WavelengthError:
            log.error("2D rectification failed!")
            log.fatal_error()
            print("Unexpected error:", sys.exc_info()[0])
            print("")
            raise


        # Automatic Background Subtraction:
        if options['skysub']['auto']:
            bgsub_pdf_name = os.path.join(output_dir, 'bgsub2D.pdf')
            log.write("Running task: Background Subtraction")
            try:
                bg_msg = auto_fit_background(rect2d_fname, bgsub2d_fname, dispaxis=1,
                                             plot_fname=bgsub_pdf_name, **options['skysub'])
                log.commit(bg_msg)
                log.write("2D sky model is saved in extension 'SKY' of the file: %s" % bgsub2d_fname)
                log.add_linebreak()
            except:
                log.error("Automatic background subtraction failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise
        else:
            log.warn("No sky-subtraction has been performed on the 2D spectrum!")
            log.write("Cosmic ray rejection may fail... double check the output or turn off 'crr' by setting niter=0.")
            log.add_linebreak()
            bgsub2d_fname = rect2d_fname


        # Correct Cosmic Rays Hits:
        if options['crr']['niter'] > 0:
            log.write("Running task: Cosmic Ray Rejection")
            crr_fname = os.path.join(output_dir, 'CRR_BGSUB2D_%s.fits' % (sci_img.target_name))
            try:
                crr_msg = correct_cosmics(bgsub2d_fname, crr_fname, **options['crr'])
                log.commit(crr_msg)
                log.add_linebreak()
            except:
                log.error("Cosmic ray correction failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise
        else:
            crr_fname = bgsub2d_fname


        # Flux Calibration:
        if status['RESPONSE']:
            log.write("Running task: Flux Calibration")
            response_fname = status['RESPONSE']
            try:
                flux_msg = flux_calibrate(crr_fname, output=flux2d_fname, response=response_fname)
                log.commit(flux_msg)
                log.add_linebreak()
                status['FLUX2D'] = flux2d_fname
            except:
                log.error("Flux calibration failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise
        else:
            status['FLUX2D'] = crr_fname


        # Extract 1D spectrum:
        log.write("Running task: 1D Extraction")
        extract_fname = status['FLUX2D']
        if options['extract']['interactive']:
            try:
                log.write("Extraction: Starting Graphical User Interface")
                extract_gui.run_gui(extract_fname, output_fname=flux1d_fname,
                                    app=app, **options['extract'])
                log.write("Writing fits table: %s" % flux1d_fname, prefix=" [OUTPUT] - ")
            except:
                log.error("Interactive 1D extraction failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise
        else:
            try:
                ext_msg = auto_extract(extract_fname, flux1d_fname, dispaxis=1, pdf_fname=extract_pdf_fname,
                                       **options['extract'])
                log.commit(ext_msg)
                log.add_linebreak()
            except:
                log.error("Spectral 1D extraction failed!")
                log.fatal_error()
                print("Unexpected error:", sys.exc_info()[0])
                raise

        log.exit()
